/****************************************************************************
** $Id: main.cpp,v 1.33 2003/10/26 09:20:42 andrew Exp $
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
**
** This file is part of the QCad project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding valid QCad Professional Edition licenses may use 
** this file in accordance with the qcadlib Commercial License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.ribbonsoft.com for further details.
**
** Contact info@ribbonsoft.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "main.h"

#include <locale.h>

#include <qapplication.h>

//#include <qwindowsstyle.h>
#ifdef __APPLE__
#include <qmacstyle_mac.h>
#endif
#include <qtimer.h>

#if QT_VERSION>=0x030200
#include <qsplashscreen.h>
#endif

#include "rs_fontlist.h"
#include "rs_patternlist.h"
#include "rs_scriptlist.h"
#include "rs_settings.h"
#include "rs_system.h"

#include "qg_dlginitial.h"

#include "qc_applicationwindow.h"

#include "xpm/splash.xpm"


/**
 * Main. Creates Application window.
 */
int main(int argc, char** argv) {
	
    //RS_DEBUG->setLevel(RS_Debug::D_DEBUGGING);
    RS_DEBUG->setLevel(RS_Debug::D_WARNING);
	RS_DEBUG->print("param 0: %s", argv[0]);

	// avoid . / , confusion on German environments
	//setlocale(LC_NUMERIC, "C");
	setlocale(LC_ALL, "C");

    QApplication app(argc, argv);

	for (int i=0; i<app.argc(); i++) {
		if (QString("--debug") == app.argv()[i]) {
    		RS_DEBUG->setLevel(RS_Debug::D_DEBUGGING);
		}
	}
	
	QFileInfo prgInfo(argv[0]);
	QString prgDir(prgInfo.dirPath(true));
    RS_SETTINGS->init(QC_COMPANYKEY, QC_APPKEY);
    RS_SYSTEM->init(QC_APPNAME, QC_VERSION, QC_APPDIR, prgDir);

	RS_String lang;
	RS_String langCmd;
	RS_String unit;
	
	RS_SETTINGS->beginGroup("/Defaults");
	unit = RS_SETTINGS->readEntry("/Unit", "Invalid");
    RS_SETTINGS->endGroup();

	// show initial config dialog:
	if (unit=="Invalid") {
		QG_DlgInitial di(NULL);
		if (di.exec()) {
			RS_SETTINGS->beginGroup("/Defaults");
			unit = RS_SETTINGS->readEntry("/Unit", "None");
    		RS_SETTINGS->endGroup();
		}
	}

#ifdef SPLASHSCREEN_H
    QPixmap* pixmap = new QPixmap(splash_xpm);
    QSplashScreen *splash = new QSplashScreen(*pixmap);
    splash->show();
	splash->message(app.tr("Loading.."), Qt::AlignRight|Qt::AlignBottom);
#endif

    RS_FONTLIST->init();
    RS_PATTERNLIST->init();
    RS_SCRIPTLIST->init();

	RS_SETTINGS->beginGroup("/Appearance");
    lang = RS_SETTINGS->readEntry("/Language", "en");
    langCmd = RS_SETTINGS->readEntry("/LanguageCmd", "en");
    RS_SETTINGS->endGroup();
	
	RS_SYSTEM->loadTranslation(lang, langCmd);

    //QApplication::setStyle(new QWindowsStyle());
    //QApplication::setStyle(new QPlatinumStyle());
#ifdef __APPLE__
	QApplication::setStyle(new QMacStyle());
#endif

#ifdef SPLASHSCREEN_H
	splash->message(app.tr("Loading Library Paths.."), 
		Qt::AlignRight|Qt::AlignBottom);
#endif
	
    QC_ApplicationWindow * mw = new QC_ApplicationWindow();
    mw->setCaption(QC_APPNAME);
    mw->show();
	mw->setFocus();

	// load given files:
	for (int i=1; i<app.argc(); i++) {
		if (QString(app.argv()[i]).startsWith("-")==false) {
			mw->slotFileOpen(app.argv()[i], RS2::FormatUnknown);
		}
	}
	
#ifdef QC_DEMO
	QTimer::singleShot(10 * 60 * 1000, mw, SLOT(slotFileDemoQuit()));
#endif
		
#ifdef SPLASHSCREEN_H
	splash->finish(mw);
    delete splash;
	delete pixmap;
#endif

    app.connect(&app, SIGNAL(lastWindowClosed()), &app, SLOT(quit()));

	// renew: avoid . / , confusion on German environments
	//setlocale(LC_NUMERIC, "C");
	setlocale(LC_ALL, "C");
    return app.exec();
}




 
