/****************************************************************************
** $Id: rs_ellipse.cpp,v 1.12 2003/11/19 00:20:27 andrew Exp $
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
**
** This file is part of the qcadlib Library project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding valid qcadlib Professional Edition licenses may use 
** this file in accordance with the qcadlib Commercial License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.ribbonsoft.com for further details.
**
** Contact info@ribbonsoft.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/


#include "rs_ellipse.h"

#include "rs_graphic.h"
#include "rs_graphicview.h"
#include "rs_painter.h"
#include "rs_information.h"

/**
 * Constructor.
 */
RS_Ellipse::RS_Ellipse(RS_EntityContainer* parent,
                       const RS_EllipseData& d)
        :RS_AtomicEntity(parent), data(d) {

    calculateEndpoints();
    calculateBorders();
}


/**
 * Recalculates the endpoints using the angles and the radius.
 */
void RS_Ellipse::calculateEndpoints() {
    double angle = data.majorP.angle();
    double radius1 = getMajorRadius();
    double radius2 = getMinorRadius();

    startpoint.set(data.center.x + cos(data.angle1) * radius1,
                   data.center.y + sin(data.angle1) * radius2);
    startpoint.rotate(data.center, angle);
    endpoint.set(data.center.x + cos(data.angle2) * radius1,
                 data.center.y + sin(data.angle2) * radius2);
    endpoint.rotate(data.center, angle);
}


/**
 * Calculates the boundary box of this ellipse.
 *
 * @todo Fix that - the algorithm used is really bad / slow.
 */
void RS_Ellipse::calculateBorders() {

    /*
    std::cout << "calc borders" << std::endl;

    double a = getMajorRadius();
    double b = getMinorRadius();
    double alpha = getAngle();


    double term = RS_Math::pow(a, 2.0) * RS_Math::pow(sin(alpha), 2.0) 
                + RS_Math::pow(b, 2.0) * RS_Math::pow(cos(alpha), 2.0);

    std::cout << "term: " << term << "\n";

    RS_Vector emax;

    if (fabs(term)>1.0e-12) {
    	emax.x = (RS_Math::pow(a, 2.0) * sin(alpha)) / sqrt(term);
    	std::cout << "emax.x: " << emax.x << "\n";
}
    else {
    	std::cout << "term is 0.0\n";
       	minV.set(0.0,0.0);
       	maxV.set(0.0,0.0);
    	return;
}

    //RS_VectorSolutions vs = mapX(emax.x);
    //emax.y = vs.vector1.y;
    emax.y = emax.x * sin(alpha) + b * sin(acos(emax.x/a)) * cos(alpha);

    std::cout << "emax.y: " << emax.y << "\n";

    maxV = data.center + emax;
    minV = data.center - emax;

    std::cout << "done" << std::endl;
    */

    double radius1 = getMajorRadius();
    double radius2 = getMinorRadius();
    double angle = getAngle();
    double a1 = ((!isReversed()) ? data.angle1 : data.angle2);
    double a2 = ((!isReversed()) ? data.angle2 : data.angle1);

    double minX = min(startpoint.x, endpoint.x);
    double minY = min(startpoint.y, endpoint.y);
    double maxX = max(startpoint.x, endpoint.x);
    double maxY = max(startpoint.y, endpoint.y);

    // kind of a brute force. TODO: calculation
    RS_Vector vp;
    for (double a=a1; a<=a2; a+=0.1) {
        vp.set(data.center.x + radius1 * cos(a),
               data.center.y + radius2 * sin(a));
        vp.rotate(data.center, angle);

        minX = min(minX, vp.x);
        minY = min(minY, vp.y);
        maxX = max(maxX, vp.x);
        maxY = max(maxY, vp.y);
    }


    minV.set(minX, minY);
    maxV.set(maxX, maxY);
    //RS_Entity::calculateBorders();
}


/**
 * Maps the given X-value to the ellipse and returns an
 * Y-value so that point (x/y) is on the ellipse.
 *
 * @todo Fix this method. Doesn't work despite complexity ;)
 */
RS_VectorSolutions RS_Ellipse::mapX(double x) {
    double a = getMajorRadius();
    double b = getMinorRadius();
    double alpha = getAngle();
    double rx = x-data.center.x;

    //std::cout << "a " << a << "\n";
    //std::cout << "b " << b << "\n";
    //std::cout << "alpha " << alpha << "\n";
    //std::cout << "rx " << rx << "\n";

    // relative x coordinates of the given point if the ellipse wasn't rotated:
    double x_n1 = 1.0/2.0/(-RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)+RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),2)-RS_Math::pow(a,2))*(-2*rx*cos(alpha)*a + 2.0*sqrt(RS_Math::pow(rx,2)*RS_Math::pow(cos(alpha),2)*RS_Math::pow(a,2)-RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)*RS_Math::pow(rx,2)+RS_Math::pow(b,4)*RS_Math::pow(sin(alpha),4)+RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),2)*RS_Math::pow(rx,2)-RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),4)*RS_Math::pow(b,2)-RS_Math::pow(a,2)*RS_Math::pow(rx,2)+RS_Math::pow(a,2)*RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)))*a;

    double x_n2 = 1.0/2.0/(-RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)+RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),2)-RS_Math::pow(a,2))*(-2*rx*cos(alpha)*a - 2.0*sqrt(RS_Math::pow(rx,2)*RS_Math::pow(cos(alpha),2)*RS_Math::pow(a,2)-RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)*RS_Math::pow(rx,2)+RS_Math::pow(b,4)*RS_Math::pow(sin(alpha),4)+RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),2)*RS_Math::pow(rx,2)-RS_Math::pow(a,2)*RS_Math::pow(sin(alpha),4)*RS_Math::pow(b,2)-RS_Math::pow(a,2)*RS_Math::pow(rx,2)+RS_Math::pow(a,2)*RS_Math::pow(b,2)*RS_Math::pow(sin(alpha),2)))*a;

    double y_n1 = sqrt(RS_Math::pow(b,2)*(1 - RS_Math::pow(x_n1,2)/RS_Math::pow(a,2)));
    double y_n2 = -sqrt(RS_Math::pow(b,2)*(1 - RS_Math::pow(x_n1,2)/RS_Math::pow(a,2)));
    double y_n3 = sqrt(RS_Math::pow(b,2)*(1 - RS_Math::pow(x_n2,2)/RS_Math::pow(a,2)));
    double y_n4 = -sqrt(RS_Math::pow(b,2)*(1 - RS_Math::pow(x_n2,2)/RS_Math::pow(a,2)));

    RS_Vector v1(false), v2(false);

    if (x_n1<=0.0 && y_n1>=0.0) {
        v1.set(x_n1, y_n1);
        v1.rotate(alpha);
        v1.move(data.center);
        std::cout << "1\n";
    }

    if (x_n1<=0.0 && y_n2<=0.0) {
        v2.set(x_n1, y_n2);
        v2.rotate(alpha);
        v2.move(data.center);
        std::cout << "2\n";
    }

    if (x_n2>=0.0 && y_n3>=0.0) {
        v1.set(x_n2, y_n3);
        v1.rotate(alpha);
        v1.move(data.center);
        std::cout << "3\n";
    }

    if (x_n2>=0.0 && y_n4<=0.0) {
        v2.set(x_n2, y_n4);
        v2.rotate(alpha);
        v2.move(data.center);
        std::cout << "4\n";
    }
    std::cout << "\n";

    //if ((alpha>=0.0 && alpha<=M_PI/2) ||
    //    (alpha>=M_PI && alpha<=3.0/2.0*M_PI)) {
    //v1.set(x_n1, y_n2);
    //v1.rotate(alpha);
    //v2.set(x_n2, y_n3);
    //v2.rotate(alpha);
    //}
    /*else {
    	v1.set(x_n1, y_n1);
    	v1.rotate(alpha);
    	v2.set(x_n2, y_n4);
    	v2.rotate(alpha);
}*/

    return RS_VectorSolutions(v1,v2);
}



RS_VectorSolutions RS_Ellipse::getRefPoints() {
	RS_VectorSolutions ret(getStartpoint(), getEndpoint(), data.center);
	return ret;
}



RS_Vector RS_Ellipse::getNearestEndpoint(const RS_Vector& coord, double* dist) {
    double dist1, dist2;
    RS_Vector* nearerPoint;

    dist1 = startpoint.distanceTo(coord);
    dist2 = endpoint.distanceTo(coord);

    if (dist2<dist1) {
        if (dist!=NULL) {
            *dist = dist2;
        }
        nearerPoint = &endpoint;
    } else {
        if (dist!=NULL) {
            *dist = dist1;
        }
        nearerPoint = &startpoint;
    }

    return *nearerPoint;
}



/**
 * @todo Implement this.
 */
RS_Vector RS_Ellipse::getNearestPointOnEntity(const RS_Vector& coord,
        bool onEntity, double* dist, RS_Entity** entity) {

    /*
    RS_Vector vec(false);
       if (entity!=NULL) {
           *entity = this;
       }

    double angle = (coord-data.center).angle();
    if (RS_Math::isAngleBetween(angle, angle1, angle2, 
                                isReversed())) {
    	vec.setPolar(radius, angle);
    	vec+=data.center;
}
    if (dist!=NULL) {
    	*dist = fabs((vec-data.center).magnitude()-radius);
    	//std::cout << " ellipse dist: " << *dist << "\n";
}

    return vec;
    */

    RS_Vector ret(false);

    if (entity!=NULL) {
        *entity = this;
    }

    double angle = data.center.angleTo(data.center+data.majorP);

    RS_Line tmpLine(NULL, RS_LineData(coord, data.center));
    tmpLine.move(-data.center);
    tmpLine.scale(RS_Vector(0.0,0.0), RS_Vector(10.0,10.0));
    tmpLine.rotate(RS_Vector(0.0,0.0), -angle);
    if (data.ratio>1.0e-6) {
        tmpLine.scale(RS_Vector(0.0,0.0), RS_Vector(1.0, 1.0/data.ratio));
    }

    /*
    RS_Graphic* graphic = getGraphic();
    RS_Entity* clone = tmpLine.clone();
    clone->reparent(graphic);
    graphic->addEntity(clone);
    */

    RS_Arc tmpArc(NULL, RS_ArcData(RS_Vector(0.0,0.0),
                                   data.center.distanceTo(data.center+data.majorP),
                                   data.angle1-angle, data.angle2-angle,
                                   //0.0, 2*M_PI,
                                   data.reversed));

    /*
    RS_Graphic* graphic = getGraphic();
    RS_Entity* clone = tmpArc.clone();
    clone->reparent(graphic);
    graphic->addEntity(clone);
    */

    RS_VectorSolutions sol =
        RS_Information::getIntersection(&tmpLine, &tmpArc, onEntity);

    ret = sol.get(0);

    ret.scale(RS_Vector(0.0,0.0), RS_Vector(1.0, data.ratio));
    ret.rotate(RS_Vector(0.0,0.0), angle);
    ret.move(data.center);

    //std::cout << "intersection: " << ret << "\n";

    if (dist!=NULL) {
        if (ret.valid) {
            *dist = ret.distanceTo(coord);
        } else {
            *dist = RS_MAXDOUBLE;
        }

        //std::cout << "dist: " << *dist << "\n";
    }

    return ret;
}



RS_Vector RS_Ellipse::getNearestCenter(const RS_Vector& coord,
                                       double* dist) {
    if (dist!=NULL) {
        *dist = coord.distanceTo(data.center);
    }
    return data.center;
}



/**
 * @todo Implement this.
 */
RS_Vector RS_Ellipse::getNearestMiddle(const RS_Vector& /*coord*/,
                                       double* dist) {
    if (dist!=NULL) {
        *dist = RS_MAXDOUBLE;
    }
    return RS_Vector(false);
}



RS_Vector RS_Ellipse::getNearestDist(double /*distance*/,
                                     const RS_Vector& /*coord*/,
                                     double* dist) {
    if (dist!=NULL) {
        *dist = RS_MAXDOUBLE;
    }
    return RS_Vector(false);
}



/**
 * @todo Implement this.
 */
double RS_Ellipse::getDistanceToPoint(const RS_Vector& coord,
                                      RS_Entity** entity,
                                      RS2::ResolveLevel) {
    double dist = RS_MAXDOUBLE;
    getNearestPointOnEntity(coord, true, &dist, entity);

    return dist;
}



void RS_Ellipse::move(RS_Vector offset) {
    data.center.move(offset);
    calculateEndpoints();
    calculateBorders();
}



void RS_Ellipse::rotate(RS_Vector center, double angle) {
    data.center.rotate(center, angle);
    data.majorP.rotate(angle);
    //data.angle1 = RS_Math::correctAngle(data.angle1+angle)i;
    //data.angle2 = RS_Math::correctAngle(data.angle2+angle);
    calculateEndpoints();
    calculateBorders();
}



void RS_Ellipse::scale(RS_Vector center, RS_Vector factor) {
    data.center.scale(center, factor);
    data.majorP.scale(factor);
    calculateEndpoints();
    calculateBorders();
}


/**
 * @todo deal with angles correctly
 */
void RS_Ellipse::mirror(RS_Vector axisPoint1, RS_Vector axisPoint2) {
    RS_Vector mp = data.center + data.majorP;

    data.center.mirror(axisPoint1, axisPoint2);
    mp.mirror(axisPoint1, axisPoint2);

    data.majorP = mp - data.center;

    double a = axisPoint1.angleTo(axisPoint2);

    RS_Vector vec;
    vec.setPolar(1.0, data.angle1);
    vec.mirror(RS_Vector(0.0,0.0), axisPoint2-axisPoint1);
    data.angle1 = vec.angle() - 2*a;

    vec.setPolar(1.0, data.angle2);
    vec.mirror(RS_Vector(0.0,0.0), axisPoint2-axisPoint1);
    data.angle2 = vec.angle() - 2*a;

    data.reversed = (!data.reversed);

    calculateEndpoints();
    calculateBorders();
}


void RS_Ellipse::draw(RS_Painter* painter, RS_GraphicView* view) {

    if (painter==NULL || view==NULL) {
        return;
    }

    double styleFactor = getStyleFactor();

    if (getPen().getLineType()==RS2::SolidLine ||
            isSelected() ||
            view->getDrawingMode()==RS2::ModePreview) {

        painter->drawEllipse(view->toGui(getCenter()),
                             getMajorRadius() * view->getFactor().x,
                             getMinorRadius() * view->getFactor().x,
                             getAngle(),
                             getAngle1(), getAngle2(),
                             isReversed());
    } else {
        // Pattern:
        RS_LineTypePattern* pat;
        if (isSelected()) {
            pat = &patternSelected;
        } else {
            pat = view->getPattern(getPen().getLineType());
        }

        if (pat==NULL) {
            return;
        }

        // Pen to draw pattern is always solid:
        RS_Pen pen = painter->getPen();
        pen.setLineType(RS2::SolidLine);
        painter->setPen(pen);

        double* da;     // array of distances in x.
        int i;          // index counter

        double length = getAngleLength();

        // create pattern:
        da = new double[pat->num];

        double tot=0.0;
        i=0;
        bool done = false;
        double curA = getAngle1();
        double curR;
        //double cx = a->getCenter().x * factor.x + offsetX;
        //double cy = - a->getCenter().y * factor.y + getHeight() - offsetY;
        RS_Vector cp = view->toGui(getCenter());
        double r1 = getMajorRadius() * view->getFactor().x;
        double r2 = getMinorRadius() * view->getFactor().x;

        do {
            curR = sqrt(RS_Math::pow(getMinorRadius()*cos(curA), 2.0)
                        + RS_Math::pow(getMajorRadius()*sin(curA), 2.0));

            if (curR>1.0e-6) {
                da[i] = fabs(pat->pattern[i] * styleFactor) / curR;
                if (pat->pattern[i] * styleFactor > 0.0) {

                    if (tot+fabs(da[i])<length) {
                        painter->drawEllipse(cp,
                                             r1, r2,
                                             getAngle(),
                                             curA,
                                             curA + da[i],
                                             false);
                    } else {
                        painter->drawEllipse(cp,
                                             r1, r2,
                                             getAngle(),
                                             curA,
                                             getAngle2(),
                                             false);
                    }
                }
            }
            curA+=da[i];
            tot+=fabs(da[i]);
            done=tot>length;

            i++;
            if (i>=pat->num) {
                i=0;
            }
        } while(!done);

        delete[] da;
    }
}



/**
 * Dumps the point's data to stdout.
 */
std::ostream& operator << (std::ostream& os, const RS_Ellipse& a) {
    os << " Ellipse: " << a.data << "\n";
    return os;
}

// EOF
